
/* Authors: Patrick Derbez, Paul Huynh, Virginie Lallemand, 
 * Maria Naya-Plasencia, Leo Perrin, Andre Schrottenloher
 * Date: feb 2020
 */

#include <iostream>
#include <iomanip>
#include <cstdint>
#include <random>
#include <ctime>

#include <exception>

#include <array>
#include <tuple>
#include <vector>

#include "primitives.hpp"
#include "crypto_aead.h"
#include "datatypes.hpp"



/**
    Returns a pair of random (x1, y1), (x1', y1') such that, regardless
    of the bundles a and b, there is a probability 2^{-24.83} that
    the pair (x1, y1, a,b) ,(x1', y1', a,b) gives a collision on the
    capacity part (two last bundles) after 4 steps (2 to 5).
*/
std::array<ClydeState,4> candidates_for_collision() {

    ClydeState w, e(1 << 22, 0, 0, 0), zero_block(0, 0, 0, 0);
    ShadowState tmp(w,w,zero_block,zero_block);
    
    ShadowState s1 = inverse_step_state(tmp, 2);

    std::array<ClydeState, 4> result;
    result[0] = s1.c[0];
    result[1] = s1.c[1];
    result[2] = s1.c[0]^e;
    result[3] = s1.c[1]^e;
    return result;
}


/**
    Checks that the candidates for collisions indeed give collisions!
*/
void check_candidates() {
    unsigned int success_count = 0;
    unsigned int nb_trials = 1 << 28;
    
    ClydeState a, ap;
    
    for (unsigned int t=0; t < nb_trials; t++)
    {
        std::array<ClydeState,4> bundles = candidates_for_collision();
        
        ShadowState mix1(bundles[0], bundles[1], a, ap);
        ShadowState mix2(bundles[2], bundles[3], a, ap);
        
        unsigned char delta_out[64] = {0}, delta_in[64] = {0},
            input_0[64] = {0}, input_1[64] = {0},
            output_0[64] = {0}, output_1[64] = {0};
        mix1.to_uc(input_0);
        mix2.to_uc(input_1);
        mix1.to_uc(output_0);
        mix2.to_uc(output_1);
        
        shadow(output_0);
        shadow(output_1);
        
        for (unsigned int j=0; j<64; j++)
        {
            delta_in[j] = input_0[j] ^ input_1[j];
            delta_out[j] = output_0[j] ^ output_1[j];
        }
        
        ShadowState out0(output_0);
        ShadowState out1(output_1);


        if ( hw(delta_out) <= 32)
        {
            success_count ++;
            std::cout << "--- Pair of inputs: " << std::endl;
            print_stream(input_0);
            print_stream(input_1);
            std::cout << "--- Input/output difference:" << std::endl;
            print_stream(delta_in);
            print_stream(delta_out);
//            std::cout << "--- Difference between bundle 0 and bundle 1 of an output:" << std::endl;
//            (out0.c[0]^out0.c[1]).print();
        }
    }
    std::cout << "prob of success:     " << std::dec << log(double(success_count)/double(nb_trials))/log(2.)  << std::endl;
    std::cout << "successfull attacks: " << std::dec << success_count  << std::endl;
}


void print_array(unsigned char *a, int len) {

    for (int i = 0; i < len/16; i++) {
         for (int j = 0; j < 16; j++) {
             std::cout << std::setw(2) << std::setfill('0') << std::hex << (int)a[(i*16)+j];
         }
         std::cout << std::endl;
    }
}


/*
    Finds collisions on the tag for the S1P mode used with 4-step Shadow.
*/
void find_collisions() {

    unsigned char k[16] = {0};
    unsigned char n[16] = {0};
    // zero nonce and key
    int success_count = 0;
    int nb_trials = 1 << 28;
    
    // now initial plaintext: encrypt zero message and get first ciphertext block
    unsigned char m[64] = {0};
    unsigned char c[80];
    unsigned long long clen = 80;
    crypto_aead_encrypt(c, &clen, m, 64, NULL, 0, NULL, n, k);
    
    for (int i = 0; i < nb_trials; i++) {
        std::array<ClydeState,4> bundles = candidates_for_collision();
        ShadowState s(bundles[0], bundles[1], bundles[2], bundles[3]);
        unsigned char tmp[64] = {0};
        s.to_uc(tmp); // just to store the bundles as unsigned chars
        // contains in order: x1', y1', x1'', y1''
        
        unsigned char m1[64] = {0};
        unsigned char m2[64] = {0};
        for (int j = 0; j < 32; j++) {
            m1[j] = c[j]^tmp[j];
            m2[j] = c[j]^tmp[j+32];
        }
        
        unsigned char c1[80], c2[80];
        crypto_aead_encrypt(c1, &clen, m1, 64, NULL, 0, NULL, n, k);
        crypto_aead_encrypt(c2, &clen, m2, 64, NULL, 0, NULL, n, k);
        
        unsigned char M1[64] = {0};
        unsigned char M2[64] = {0};
        for (int j = 0; j < 32; j++) {
            M1[j] = m1[j]; // c[j]^tmp[j];
            M2[j] = m2[j];// c[j]^tmp[j+32];
            M1[j+32] = c1[j+32];
            M2[j+32] = c2[j+32];
        }
        
        crypto_aead_encrypt(c1, &clen, M1, 64, NULL, 0, NULL, n, k);
        crypto_aead_encrypt(c2, &clen, M2, 64, NULL, 0, NULL, n, k);
        bool ok = true;
        for (int j = 0; j < 16; j++) {
            if (c1[64+j] != c2[64+j]) {
                ok = false;
            }
        }
        if (ok) {
            success_count++;
            std::cout << "Found! Input messages:" << std::endl;
            print_array(M1, 64);
            print_array(M2, 64);
            std::cout << "Outputs:" << std::endl;
            print_array(c1, 80);
            print_array(c2, 80);
        }
    }
    
    std::cout << "prob of success:     " << std::dec << log(double(success_count)/double(nb_trials))/log(2.)  << std::endl;
    std::cout << "successfull attacks: " << std::dec << success_count  << std::endl;
    
    
}


int main() {

    std::cout << "SHADOW_NBYTES is set to " << SHADOW_NBYTES << std::endl;
    std::cout << "SHADOW_NS is set to " << SHADOW_NS << std::endl;
    std::cout << "SHADOW_START is set to " << SHADOW_START << std::endl;
    std::cout << "SHADOW_NR (number of rounds) is set to " << SHADOW_NR << std::endl;
    std::cout << "SHADOW_STARTR (starting round) is set to " << SHADOW_STARTR << std::endl;
    
//    check_candidates();

    find_collisions();

    return 0;
}





