
/* Authors: Patrick Derbez, Paul Huynh, Virginie Lallemand, 
 * Maria Naya-Plasencia, Leo Perrin, Andre Schrottenloher
 * Date: feb 2020
 */


#include <iostream>
#include <iomanip>
#include <cstdint>
#include <random>
#include <ctime>

#include <exception>

#include <array>
#include <tuple>
#include <vector>
#include "primitives.hpp"
#include "datatypes.hpp"


void print_bin(uint32_t x)
{
  for(unsigned int col=0; col<32; col++)
    {
      std::cout << std::dec << ((x >> col) & 1) ;
      if (col % 8 == 7)
        std::cout << " ";
    }
}


std::mt19937 ClydeState::PRNG = std::mt19937(123);

/* !SECTION! ClydeState members */
/* ====================== */

ClydeState::ClydeState() // initialized to a random content by default
{
    std::uniform_int_distribution<uint32_t> dist;
    for (unsigned int i=0; i<4; i++)
        x[i] = dist(PRNG);
}

ClydeState::ClydeState(uint32_t x0, uint32_t x1, uint32_t x2, uint32_t x3)
{
    x[0] = x0; x[1] = x1;
    x[2] = x2; x[3] = x3;
}

ClydeState::ClydeState(const unsigned char* bytes) {
    uint32_t a[4];
    bytes2state(a, bytes);
    x[0] = a[0];x[1] = a[1];x[2] = a[2];x[3] = a[3];
}

ClydeState::ClydeState(const ClydeState &c) : x(c.x) {}


void ClydeState::print()
{
  std::cout << "\n";
  for(unsigned row=0; row<4; row++)
    {
      print_bin(x[row]);
      std::cout << "\n";
    }
    // std::cout << "\n";
    // for(unsigned int i=0; i<4; i+=1)
    //     std::cout << std::hex << std::setw(8) << std::setfill('0')
    //               << x[i] << "\n";
}


int ClydeState::hw() {
    int result = 0;
    for (int i = 0; i < 4; i++) {
        for(unsigned int col=0; col<32; col++)
        {
            result += ((x[i] >> col) & 1);
        }
    }
    return result;
}

ClydeState ClydeState::operator^(const ClydeState &c) const
{
    ClydeState result(c);
    for (unsigned int i=0; i<4; i++)
        result.x[i] ^= x[i];
    return result;
}

bool ClydeState::operator==(const ClydeState &c) const
{
    for (unsigned int i=0; i<4; i++)
        if (c.x[i] != x[i])
            return false;
    return true;
}

bool ClydeState::operator!=(const ClydeState &c) const
{
    return !(*this == c);
}





/* !SECTION! ShadowState members */
/* ====================== */

ShadowState::ShadowState() {}

ShadowState::ShadowState(const unsigned char* bytes) {
    for (int b = 0; b < MLS_BUNDLES; b++) {
        const unsigned char* ptr = bytes + (b * SHADOW_NBYTES / MLS_BUNDLES);
        c[b] = ClydeState(ptr);
    }
}

#if SMALL_PERM
ShadowState::ShadowState(ClydeState c0, ClydeState c1, ClydeState c2) :
c({c0, c1, c2}) {}
#else
ShadowState::ShadowState(ClydeState c0, ClydeState c1, ClydeState c2, ClydeState c3) :
    c({c0, c1, c2, c3}) {}
#endif

ShadowState::ShadowState(const ShadowState &s) : c(s.c) {}

ShadowState ShadowState::operator^(const ShadowState &s) const
{
    ShadowState result(s);
    for (unsigned int i=0; i<MLS_BUNDLES; i++)
        result.c[i] = result.c[i] ^ c[i];
    return result;
}

bool ShadowState::operator==(const ShadowState &s) const
{
    for (unsigned int i=0; i<MLS_BUNDLES; i++)
        if (s.c[i] != c[i])
            return false;
    return true;
}

bool ShadowState::operator!=(const ShadowState &s) const
{
    return !(*this == s);
}

void ShadowState::print()
{
    for(unsigned int i=0; i<4; i++)
    {
        for (unsigned int j=0; j<MLS_BUNDLES; j++)
        {
            std::cout << std::hex << std::setw(8) << std::setfill('0')
                      << c[j].x[i] << " ";
        }
        std::cout << "\n";
    }
}

void ShadowState::to_uc(unsigned char result[])
{
    for (unsigned int b = 0; b < MLS_BUNDLES; b++)
    {
        uint32_t state[4] = {c[b].x[0], c[b].x[1], c[b].x[2], c[b].x[3]} ;
        state2bytes(&result[0] + (b * SHADOW_NBYTES / MLS_BUNDLES), state);
    }
}



/* !SECTION! Sub-routines */
/* ====================== */

void print_stream(unsigned char * bytes)
{
    for (unsigned int i=0; i<4; i++)
    {
        for (unsigned int k=0; k<MLS_BUNDLES; k++)
        {
            for (unsigned int l=0; l<4; l++)
            {
                // unsigned char p = bytes[(k*16)+(i*4)+3-l];
                // for (unsigned int b=0; b<8; b++)
                //     std::cout << ((p >> (7-b)) & 1) ;
                std::cout << std::setw(2) << std::setfill('0') << std::hex << (int)bytes[(k*16)+(i*4)+3-l];
            }
            std::cout << " ";
        }
        std::cout << std::endl;
    }
}


/* !SUBSECTION! Wrappers for clydes/bundles */

ClydeState sbox_layer_clyde(ClydeState c)
{
    ClydeState result(c);
    sbox_layer(&result.x[0]);
    return result;
}


ClydeState sbox_layer_inv_clyde(ClydeState c)
{
    ClydeState result(c);
    sbox_layer_inv(&result.x[0]);
    return result;
}


ClydeState lbox_layer_clyde(ClydeState c)
{
    ClydeState result(c);
    lbox_layer(&result.x[0]);
    return result;
}


ClydeState lbox_layer_inv_clyde(ClydeState c)
{
    ClydeState result(c);
    lbox_layer_inv(&result.x[0]);
    return result;
}


ClydeState super_sbox(ClydeState c,
                      unsigned int s, // step index
                      unsigned int b) // "mls bundle" (super-sbox) index
{
    ClydeState result(c);
    result = sbox_layer_clyde(result);
    result = lbox_layer_clyde(result);
    add_rc(&result.x[0], 2 * s, b);
    result = sbox_layer_clyde(result);
    return result;
}


ClydeState super_sbox_inv(ClydeState c,
                          unsigned int s, // step index
                          unsigned int b) // "mls bundle" (super-sbox) index
{
    ClydeState result(c);
    result = sbox_layer_inv_clyde(result);
    add_rc(&result.x[0], 2 * s, b);
    result = lbox_layer_inv_clyde(result);
    result = sbox_layer_inv_clyde(result);
    return result;
}


/* !SUBSECTION! Wrappers for full ShadowState */

ShadowState add_rc_shadow(ShadowState state,
                          unsigned int s) // step index
{
    ShadowState result(state);
    for (unsigned int b = 0; b < MLS_BUNDLES; b++)
        add_rc(&result.c[b].x[0], 2*s+1, b);
    return result;
}


/**
    Applies the Shadow D-box layer.
*/
ShadowState dbox_layer(ShadowState s)
{
    ShadowState result(s);
    #if SMALL_PERM
    ClydeState x(result.c[0]), y(result.c[1]), z(result.c[2]);
    result.c[0] = x^y^z;
    result.c[1] = x^z;
    result.c[2] = x^y;
    #else
    ClydeState sum(result.c[0] ^ result.c[1] ^ result.c[2] ^ result.c[3]);
    for (unsigned int i=0; i<MLS_BUNDLES; i++)
        result.c[i] = result.c[i] ^ sum;
    #endif
    return result;
}


/**
    Applies the Shadow inverse D-box layer. /!\ The D-box is not involutory
    in Shadow-384, contrary to Shadow-512.
*/
ShadowState inverse_dbox_layer(ShadowState s)
{
    ShadowState result(s);
    #if SMALL_PERM
    ClydeState x(result.c[0]), y(result.c[1]), z(result.c[2]);
    result.c[0] = x^y^z;
    result.c[1] = x^y;
    result.c[2] = x^z;
    #else
    ClydeState sum(result.c[0] ^ result.c[1] ^ result.c[2] ^ result.c[3]);
    for (unsigned int i=0; i<MLS_BUNDLES; i++)
        result.c[i] = result.c[i] ^ sum;
    #endif
    return result;
}


/**
    Applies step number s.
*/
ShadowState step_state(ShadowState start, unsigned int s)
{
    ShadowState state(start);
    for (unsigned int b=0; b<MLS_BUNDLES; b++)
        state.c[b] = super_sbox(state.c[b], s, b);
    state = dbox_layer(state);
    state = add_rc_shadow(state, s);
    return state;
}


/**
    Applies the inverse of step number s.
*/
ShadowState inverse_step_state(ShadowState start, unsigned int s)
{
    ShadowState state = add_rc_shadow(start, s);
    state = inverse_dbox_layer(state);
    for (unsigned int b=0; b<MLS_BUNDLES; b++)
        state.c[b] = super_sbox_inv(state.c[b], s, b);
    return state;
}


/**
    Applies the inverse of *round* number r.
*/
ShadowState inverse_round_state(ShadowState start, unsigned int r)
{
    ShadowState state = start;
    if (r % 2 == 1) {
        for (unsigned int b=0; b<MLS_BUNDLES; b++) {
            add_rc(&(state.c[b]).x[0], r, b);
        }
//        state = add_rc_shadow(start, int(r / 2));
        state = inverse_dbox_layer(state);
        for (unsigned int b=0; b<MLS_BUNDLES; b++) {
            state.c[b] = sbox_layer_inv_clyde(state.c[b]);
        }
    }
    else {
        state = start;
        for (unsigned int b=0; b<MLS_BUNDLES; b++) {
            add_rc(&(state.c[b]).x[0], r, b);
            state.c[b] = lbox_layer_inv_clyde(state.c[b]);
            state.c[b] = sbox_layer_inv_clyde(state.c[b]);
        }
    }
    return state;
}


/**
    Returns the number of nonzero bytes in a byte array of SHADOW_NBYTES size.
*/
unsigned int hw(unsigned char * bytes)
{
    unsigned int result = 0;
    for (unsigned int j=0; j<SHADOW_NBYTES; j++)
        if (bytes[j] > 0)
            result ++;
    return result;
}


/**
    Tests our wrappers for Clyde and Shadow components (either 384-bit
    or 512-bit, wheter SMALL_PERM is set to 0 or 1).
*/
void test_shadow() {
    
    std::cout << "Testing our Shadow wrapper..." << std::endl;
    std::cout << "Shadow NBYTES is set to " << SHADOW_NBYTES << std::endl;
    int nb_tests = 1 << 10;
    for (int i = 0; i < nb_tests; i++) {
        ShadowState s; // random state by default
        
        unsigned char input[SHADOW_NBYTES] = {0}, output[SHADOW_NBYTES] = {0},
                      inputcopy[SHADOW_NBYTES] = {0};
        s.to_uc(input);
        s.to_uc(inputcopy);
        for (int j = 0; j < 6; j++) {
            s = step_state(s,j);
        }
        shadow(input);
        s.to_uc(output);
        
        for (unsigned int j=0; j<SHADOW_NBYTES; j++) {
            if (input[j] != output[j]) {
                std::cout << "Test failed! Outputs are:" << std::endl;
                print_stream(input);
                std::cout << "---" << std::endl;
                print_stream(output);
                throw std::runtime_error("Test failed!");
            }
        }
        
        // now try the other way around: pass the genuine shadow output
        // state into our inverse permutation
        s = ShadowState(input);
        for (int j = 5; j >= 0; j--) {
            s = inverse_step_state(s,j);
        }
        s.to_uc(output);
        
        for (unsigned int j=0; j<SHADOW_NBYTES; j++) {
            if (output[j] != inputcopy[j]) {
                std::cout << "Test failed! (in the reverse direction) Outputs are:" << std::endl;
                print_stream(input);
                std::cout << "---" << std::endl;
                print_stream(output);
                throw std::runtime_error("Test failed!");
            }
        }
        
        
    }
    std::cout << "Success!" << std::endl;

}


