
/* Authors: Patrick Derbez, Paul Huynh, Virginie Lallemand, 
 * Maria Naya-Plasencia, Leo Perrin, Andre Schrottenloher
 * Date: feb 2020
 */

#include <iostream>
#include <iomanip>
//#include <cstdint>
//#include <random>
#include <ctime>

#include <exception>

//#include <array>
#include <tuple>
#include <vector>

#include "primitives.hpp"
#include "datatypes.hpp"

/**
    Finds the probability that a pair x y y , x z z gives same pattern after step i
*/
void check_pattern() {
    unsigned int nb_trials = 1 << 22;
    unsigned int success_count = 0;
    
    for (int nb = 3; nb < 5; nb++) 
    {
        success_count = 0;
        for (unsigned int t=0; t < nb_trials; t++) 
        {
            ClydeState x,y,yp, zero_block(0, 0, 0, 0);
            ShadowState start(x, y, y);
            ShadowState startp(x, yp, yp);
            
            ShadowState end = step_state(start, nb);
            ShadowState endp = step_state(startp, nb);
            
            if ((end.c[1] == end.c[2])
                 and (endp.c[1] == endp.c[2]) 
                 and (end.c[0] == endp.c[0]) )
            {
                success_count ++;
            }
        }
        
    std::cout << "Step: " << nb << std::endl;
    std::cout << "prob of success:     " << std::dec << 
        log(double(success_count)/double(nb_trials))/log(2.)  << std::endl;
    }
}

/**
    Finds the probability that a pair x y y , x z z
    gives x * *, x * * after step i
*/
void check_pattern_final() {
    unsigned int nb_trials = 1 << 22;
    unsigned int success_count = 0;
    
    for (int nb = 5; nb < 6; nb++) 
    {
        success_count = 0;
        for (unsigned int t=0; t < nb_trials; t++) 
        {
            ClydeState x,y,yp, zero_block(0, 0, 0, 0);
            ShadowState start(x, y, y);
            ShadowState startp(x, yp, yp);
            
            ShadowState end = step_state(start, nb);
            ShadowState endp = step_state(startp, nb);
            
            if (end.c[0] == endp.c[0]) 
            {
                success_count ++;
            }
        }
        
    std::cout << "Step: " << nb << std::endl;
    std::cout << "prob of success:     " << std::dec << 
        log(double(success_count)/double(nb_trials))/log(2.)  << std::endl;
    }
}


/**
    Finds pairs of states conforming to our 6-round path from round 1 to round 6.
*/
void shadow384_6rounds_distinguisher(bool verbose)
{
    unsigned int success_count = 0;
    unsigned int nb_trials = 1 << 30;
    
    for (unsigned int t=0; t < nb_trials; t++)
    {
        ClydeState z,y,yp,e(1 << 22, 0, 0, 0), zero_block(0, 0, 0, 0);
        ShadowState end(z, y, y);

        ShadowState s1 = inverse_step_state(end, 2);
        ShadowState s2(s1.c[0], s1.c[1]^e, s1.c[2]^e);

        s1 = inverse_step_state(s1, 1);
        s2 = inverse_step_state(s2, 1);

        unsigned char delta_out[48] = {0}, delta_in[48] = {0},
            input_0[48] = {0},
            input_1[48] = {0},
            output_0[48] = {0},
            output_1[48] = {0};
        s1.to_uc(input_0);
        s2.to_uc(input_1);
        
        s1.to_uc(output_0);
        s2.to_uc(output_1);
        shadow(output_0);
        shadow(output_1);
        
        ShadowState out0(output_0);
        ShadowState out1(output_1);
        // invert the final D layer
        out0 = inverse_dbox_layer(out0);
        out1 = inverse_dbox_layer(out1);
        out0.to_uc(output_0);
        out1.to_uc(output_1);
        
        for (unsigned int j=0; j<48; j++)
        {
            delta_in[j] = input_0[j] ^ input_1[j];
            delta_out[j] = output_0[j] ^ output_1[j];
        }
        
        if (hw(delta_out) <= 32)
        {
            success_count++;
            if (verbose)
            {
                std::cout << "--- Pair of inputs: " << std::endl;
                print_stream(input_0);
                print_stream(input_1);
                std::cout << "--- Input/output difference:" << std::endl;
                print_stream(delta_in);
                print_stream(delta_out);
            }
        }
    }

    std::cout << "prob of success:     " << std::dec << log(double(success_count)/double(nb_trials))/log(2.)  << std::endl;
    std::cout << "successfull attacks: " << std::dec << success_count  << std::endl;

}


int main() {

    std::cout << "We now find input-output pairs for our 6-round Shadow-384 distinguisher" << std::endl;
    std::cout << "SHADOW_NBYTES is set to " << SHADOW_NBYTES << std::endl;
    std::cout << "SHADOW_NS is set to " << SHADOW_NS << std::endl;
    std::cout << "SHADOW_START is set to " << SHADOW_START << std::endl;
    
    // set the "verbose" parameter to false if you want to compute a precise
    // probability of success, and not to display examples
    shadow384_6rounds_distinguisher(true);

    return 0;
}


