
/* Authors: Patrick Derbez, Paul Huynh, Virginie Lallemand, 
 * Maria Naya-Plasencia, Leo Perrin, Andre Schrottenloher
 * Date: feb 2020
 */

#include <iostream>
#include <iomanip>
#include <cstdint>
#include <random>
#include <ctime>

#include <exception>

#include <array>
#include <tuple>
#include <vector>

#include "primitives.hpp"
#include "datatypes.hpp"


void shadow512_7rounds_distinguisher(bool verbose) 
{

    unsigned int success_count = 0, nb_trials = 1 << 22;
    for (unsigned int t=0; t < nb_trials; t++) 
    {
        ClydeState y, z, zero_block(0, 0, 0, 0);
        ShadowState after_step_2 (y, y, y, z);

        // Here we can choose any combination of differences in bits
        // 22 and 23
        // Below are a few examples
        ClydeState epsilon((1 << 22) + (1 << 23), 0, 0, 0);
        // ClydeState epsilon(1 << 22, 0, 0, 1 << 23);
        
        ShadowState difference_epsilon(epsilon, epsilon, epsilon, zero_block),
            state_input_0 = inverse_step_state(after_step_2, 2),
            state_input_1 = state_input_0 ^ difference_epsilon;
        
        state_input_0 = inverse_step_state(state_input_0, 1);
        state_input_0 = inverse_step_state(state_input_0, 0);
        state_input_1 = inverse_step_state(state_input_1, 1);
        state_input_1 = inverse_step_state(state_input_1, 0);
        
        unsigned char
            input_0[64] = {0}, input_1[64] = {0},
            output_0[64] = {0}, output_1[64] = {0},
            delta_out[64] = {0}, delta_in[64] = {0};

        state_input_0.to_uc(input_0);
        state_input_1.to_uc(input_1);
       
        // compute Shadow using the reference implementation
        for (unsigned int j=0; j<64; j++)
        {
            output_0[j] = input_0[j];
            output_1[j] = input_1[j];
        }
        shadow(output_0);
        shadow(output_1);
        // uncompute the last D-box
        ShadowState tmp(output_0);
        tmp = inverse_dbox_layer(tmp);
        tmp.to_uc(output_0);
        
        tmp = ShadowState(output_1);
        tmp = inverse_dbox_layer(tmp);
        tmp.to_uc(output_1);
        
        // compute the input and output differences
        for (unsigned int j=0; j<64; j++)
        {
            delta_in[j] = input_0[j] ^ input_1[j];
            delta_out[j] = output_0[j] ^ output_1[j];
        }

        // if there are many zeroes in output, we found a pair!
        unsigned int hw_delta_out = hw(delta_out);
        if (hw_delta_out <= 48)
        {
            if (verbose)
            {
                std::cout << "--- Pair of inputs: " << std::endl;
                print_stream(input_0);
                print_stream(input_1);
                std::cout << "--- Input/output difference:" << std::endl;
                print_stream(delta_in);
                print_stream(delta_out);
            }
            success_count ++;
        }
    }
    
    std::cout << "prob of success:     " << std::dec << log(double(success_count)/double(nb_trials))/log(2.)  << std::endl;
    std::cout << "successfull attacks: " << std::dec << success_count  << std::endl;

}


int main() {

    std::cout << "We now find input-output pairs for our 7-round Shadow-512 distinguisher" << std::endl;
    std::cout << "SHADOW_NBYTES is set to " << SHADOW_NBYTES << std::endl;
    std::cout << "SHADOW_NS is set to " << SHADOW_NS << std::endl;
    
    // set the "verbose" parameter to false if you want to compute a precise
    // probability of success, and not to display examples
    shadow512_7rounds_distinguisher(true);

    return 0;
}

